
/* A5SES is released under the zlib license.
 * See LICENSE.txt for more info.
 */

/// @file Event.hpp
/// @ingroup Event
/// Contains generic event queue related classes

#ifndef A5SES_EVENT_HPP_INCLUDED
#define A5SES_EVENT_HPP_INCLUDED

#include <list>
#include <memory>

#include "common.hpp"

namespace a5
{

class Event_Queue;
class Event;

/// @ingroup Event
/// An Event_Source can be registered to an Event_Queue
class Event_Source
{
	protected:
		/// Pointer to the Allegro event source
		ALLEGRO_EVENT_SOURCE *event_source;

	public:
		static const int Any = 0xFFFFFFFF;

		/// Processes Allegro event messages in to A5SES Events
		virtual std::auto_ptr<Event> Handle(ALLEGRO_EVENT *raw_event) const
		{
			return std::auto_ptr<Event>(0);
		}

	friend class Event_Queue;
};

/// @ingroup Event
/// Base class for event messages
class Event
{
	protected:
		/// The subtype of the event
		int subtype;

	public:
		/// The default subtype
		static const int Invalid = -1;

		/// Pointer to the Event_Source that created the event
		const Event_Source *source;

		/// A globally unique ID number for this event type
		static const int TypeID = 0;

		/// Returns the TypeID of the event
		virtual int Type()
		{
			return TypeID;
		}

		/// Returns the subtype of the event
		int SubType()
		{
			return this->subtype;
		}
};

/// @ingroup Event
/// Processes events from Event_Sources
class Event_Queue
{
	protected:
		/// Pointer to the Allegro event queue
		ALLEGRO_EVENT_QUEUE *event_queue;

		/// List of pointers to registered event sources
		std::list<const Event_Source *> sources;

	public:
		/// Creates a new event queue
		Event_Queue()
		{
			_init();
			this->event_queue = al_create_event_queue();
		}

		/// Checks a raw ALLEGRO_EVENT through the event source handlers
		/// Returns NULL if no source handlers caused the event
		std::auto_ptr<Event> Handle(ALLEGRO_EVENT *raw_event)
		{
			std::auto_ptr<Event> event;

			for (std::list<const Event_Source *>::iterator i = this->sources.begin(); i != this->sources.end(); ++i)
			{
				if ((event = (*i)->Handle(raw_event)).get())
				{
					break;
				}
			}

			return event;
		}

		/// Waits forever for an event
		std::auto_ptr<Event> Wait()
		{
			static ALLEGRO_EVENT raw_event;

			al_wait_for_event(this->event_queue, &raw_event);
			std::auto_ptr<Event> event = this->Handle(&raw_event);

			if (event.get())
			{
				return event;
			}
			else
			{
				return this->Wait();
			}
		}

		/// Waits for an event for a specified number of seconds
		/// Returns NULL if the timer expires
		std::auto_ptr<Event> Wait(float seconds)
		{
			static ALLEGRO_EVENT event;

			if (al_wait_for_event_timed(this->event_queue, &event, seconds))
			{
				return this->Handle(&event);
			}
			else
			{
				return std::auto_ptr<Event>(0);
			}
		}

		/// Checks for and returns an event. Returns NULL if the queue is empty
		std::auto_ptr<Event> Get()
		{
			static ALLEGRO_EVENT event;

			if (al_get_next_event(this->event_queue, &event))
			{
				return this->Handle(&event);
			}
			else
			{
				return std::auto_ptr<Event>(0);
			}
		}

		/// Checks for and returns an event. The event remains in the queue.
		/// Returns NULL if the queue is empty
		std::auto_ptr<Event> Peek()
		{
			static ALLEGRO_EVENT event;

			if (al_peek_next_event(this->event_queue, &event))
			{
				return this->Handle(&event);
			}
			else
			{
				return std::auto_ptr<Event>(0);
			}
		}

		/// Returns true if the event queue is empty
		bool Empty()
		{
			return al_event_queue_is_empty(this->event_queue);
		}

		/// Removes all events from the queue
		void Flush()
		{
			al_flush_event_queue(this->event_queue);
		}

		/// Removes the next even from the queue. Returns true if an event was dropped
		bool Drop()
		{
			return al_drop_next_event(this->event_queue);
		}

		/// Registers an event source
		void Register(const Event_Source &source)
		{
			this->sources.push_back(&source);
			al_register_event_source(this->event_queue, source.event_source);
		}

		/// Unregisters an event source
		void Unregister(const Event_Source &source)
		{
			this->sources.remove(&source);
			al_unregister_event_source(this->event_queue, source.event_source);
		}

		/// Unregisters any event sources and frees resources
		~Event_Queue()
		{
			for (std::list<const Event_Source *>::iterator i = this->sources.begin(); i != this->sources.end(); ++i)
			{
				al_unregister_event_source(this->event_queue, (*i)->event_source);
			}

			al_destroy_event_queue(this->event_queue);
		}
};

}

#endif // A5SES_EVENT_HPP_INCLUDED
