
/* A5SES is released under the zlib license.
 * See LICENSE.txt for more info.
 */

/// @file Mouse.hpp
/// @ingroup Mouse
/// Contains mouse related classes

#ifndef A5SES_MOUSE_HPP_INCLUDED
#define A5SES_MOUSE_HPP_INCLUDED

#include "Event.hpp"

namespace a5
{

/// @internal
/// @ingroup Internal Mouse
/// Initializes the Allegro mouse library
inline void _mouse_init()
{
	if (!al_is_mouse_installed())
	{
		_init();

		if (!(al_install_mouse()))
		{
			throw Init_Failed("mouse");
		}
	}
}

/// @ingroup Mouse
/// Object representing a mouse
class Mouse : public Event_Source
{
	public:
		enum Button
		{
			Invalid = 0,
			Left = 1,
			Right = 2,
			Middle = 3,
		};

		enum Axis
		{
			X,
			Y,
			Z,
			W
		};

		/// @ingroup Event Mouse
		/// Mouse event object
		class Event : public a5::Event
		{
			public:
				/// Event subtypes
				enum Sub
				{
					Move = 0x00000001,
					Down = 0x00000002,
					Up = 0x00000004,
					Warp = 0x00000008,
					Enter = 0x00000010,
					Leave = 0x00000020,
					Any = 0xFFFFFFFF,
				};

				static const int TypeID = 0x00000002;

				/// Current position of the horizontal scroll
				int w;

				/// Current X coordinate of the mouse cursor
				int x;

				/// Current Y coordinate of the mouse cursor
				int y;

				/// Current position of the vertical scroll
				int z;

				/// Difference this event caused to the position of the horizontal scroll
				int dw;

				/// Difference this event caused to the X coordinate of the mouse cursor
				int dx;

				/// Difference this event caused to the Y coordinate of the mouse cursor
				int dy;

				/// Difference this event caused to the position of the vertical scroll
				int dz;

				/// Button that was clicked
				Button button;

				/// Creates a Mouse event
				Event(const Event_Source *source_, int subtype_, int w_, int x_, int y_, int z_, int dw_, int dx_, int dy_, int dz_, Button button_) : w(w_), x(x_), y(y_), z(z_), dw(dw), dx(dx_), dy(dy_), dz(dz_), button(button_)
				{
					this->source = source_;
					this->subtype = subtype_;
				}

				virtual int Type()
				{
					return TypeID;
				}
		};

		/// @ingroup Event Mouse
		/// Mouse state
		class State
		{
			public:
				/// Position of the horizontal scroll
				int w;

				/// X coordinate of the mouse cursor
				int x;

				/// Y coordinate of the mouse cursor
				int y;

				/// Position of the vertical scroll
				int z;

				/// Bitmask of buttons held down
				int buttons;

				/// Creates a Mouse State from an Allegro mouse state
				State(const ALLEGRO_MOUSE_STATE &state) : w(state.w), x(state.x), y(state.y), z(state.z), buttons(state.buttons) {}

				/// Returns the Allegro mouse state
				operator ALLEGRO_MOUSE_STATE()
				{
					ALLEGRO_MOUSE_STATE state;

					state.w = this->w;
					state.x = this->x;
					state.y = this->y;
					state.z = this->z;
					state.buttons = this->buttons;

					return state;
				}
		};

		/// @ingroup Event Mouse
		/// Mouse cursor
		class Cursor
		{
			protected:
				/// Pointer to Allegro mouse cursor
				ALLEGRO_MOUSE_CURSOR *cursor;

			public:
				/// Creates a mouse cursor from a Bitmap
				Cursor(Bitmap &bmp, int focus_x, int focus_y)
				{
					this->cursor = al_create_mouse_cursor(bmp, focus_x, focus_y);
				}

				/// Destroys the mouse cursor
				~Cursor()
				{
					al_destroy_mouse_cursor(this->cursor);
				}

				/// Returns a pointer to the Allegro mouse cursor
				operator ALLEGRO_MOUSE_CURSOR *()
				{
					return this->cursor;
				}
		};

		/// Creates a Mouse
		Mouse()
		{
			_mouse_init();
			this->event_source = reinterpret_cast<ALLEGRO_EVENT_SOURCE *>(al_get_mouse());
			//this->event_source = al_get_mouse_event_source(al_get_mouse());
		}

		/// Takes a snapshot of the current state of the mouse
		a5::Mouse::State GetState()
		{
			ALLEGRO_MOUSE_STATE state;

			al_get_mouse_state(&state);

			return a5::Mouse::State(state);
		}

		/// Restores the coordinates to a previous State
		void SetState(a5::Mouse::State state)
		{

		}

		/// Changes the range of accepted mouse values
		bool SetRange(Rectangle r)
		{
			return al_set_mouse_range(r.x1, r.y1, r.x2, r.y2);
		}

		virtual std::auto_ptr<a5::Event> Handle(ALLEGRO_EVENT *raw_event) const
		{
			std::auto_ptr<a5::Event> event(0);

			if (raw_event->any.source == this->event_source)
			{
				int subtype = a5::Event::Invalid;

				switch (raw_event->type)
				{
					case ALLEGRO_EVENT_MOUSE_AXES:
						subtype = Event::Move;
						break;

					case ALLEGRO_EVENT_MOUSE_BUTTON_DOWN:
						subtype = Event::Down;
						break;

					case ALLEGRO_EVENT_MOUSE_BUTTON_UP:
						subtype = Event::Up;
						break;

					case ALLEGRO_EVENT_MOUSE_WARPED:
						subtype = Event::Warp;
						break;

					case ALLEGRO_EVENT_MOUSE_ENTER_DISPLAY:
						subtype = Event::Enter;
						break;

					case ALLEGRO_EVENT_MOUSE_LEAVE_DISPLAY:
						subtype = Event::Leave;
						break;
				}

				event = std::auto_ptr<a5::Event>(new Event(this, subtype, raw_event->mouse.w, raw_event->mouse.x, raw_event->mouse.y, raw_event->mouse.z, raw_event->mouse.dw, raw_event->mouse.dx, raw_event->mouse.dy, raw_event->mouse.dz, static_cast<Button>(raw_event->mouse.button)));
			}

			return event;
		}
};

}

#endif // A5SES_MOUSE_HPP_INCLUDED
